
# Main Sections:
#     Changing the tree topology: adding nodes, reparenting nodes
#     Tree traversal
#     Filling in costs of veins: weights, bends, accumulation
#     Making grid versions of the tree for display and growth
#     Vein optimisation
#     Adding cells at border
#     Display, as image and text
#     Top level 

image.dir = "C:/Users/GRJ/AAA/Programming/Biology/GrowingLeaves/article/simple/"


################################################################################
# Changing the tree topology: adding nodes, reparenting nodes

make.root <- function() {
	root = list(N=1, pN=-1, cNs=integer(0), x=0, y=0, w=NA, m=1, f=1, F=1)
	list(root)
}


make.node <- function() {
	list(N=-1, pN=-1, cNs=integer(0), x=NA, y=NA, w=NA, m=NA, f=NA, F=NA)
}


add.child <- function(tree, pN, x, y) {
	node = make.node()
	node$N = length(tree) + 1
	node$pN = pN
	node$x = x
	node$y = y
	tree[[pN]]$cNs = c(tree[[pN]]$cNs, node$N)
	new.tree = c(tree , list(node))
	new.tree
}


# can reparent to existing parent (does nothing)
reparent.node <- function(tree, n, qN) {
	# remove n from current parent's children
	pN = tree[[n]]$pN
	pcNs = tree[[pN]]$cNs
	tree[[pN]]$cNs = pcNs[pcNs != n]
	# make qN the new parent
	tree[[n]]$pN = qN
	tree[[qN]]$cNs = c(tree[[qN]]$cNs, n)
	tree
}


################################################################################
# Tree traversal

collect.subtree.postorder <- function(tree, N, Ns) {
	node = tree[[N]]
	nc = length(node$cNs)
	if (nc > 0) {
		for (ch in 1:nc) {
			Ns = collect.subtree.postorder(tree, node$cNs[ch], Ns)
		}
	}
	Ns = c(Ns, node$N)
	Ns
}


collect.subtree.preorder <- function(tree, N, Ns) {
	node = tree[[N]]
	Ns = c(Ns, node$N)
	nc = length(node$cNs)
	if (nc > 0) {
		for (ch in 1:nc) {
			Ns = collect.subtree.preorder(tree, node$cNs[ch], Ns)
		}
	}
	Ns
}

collect.postorder <- function(tree) {
	Ns = NULL
	Ns = collect.subtree.postorder(tree, 1, Ns)
	Ns
}  


collect.preorder <- function(tree) {
	Ns = NULL
	Ns = collect.subtree.preorder(tree, 1, Ns)
	Ns
}  


################################################################################
# Utility functions

dist.between.nodes <- function(tree, node, node2) {
	x = node$x
	y = node$y
	px = node2$x
	py = node2$y
	dx = x - px
	dy = y - py
	dist = sqrt(dx*dx+dy*dy)	
	dist
}


dist.to.node <- function(tree, node) {
	pN = node$pN
	dist = 1 # for root
	if (pN > 0) {
		dist = dist.between.nodes(tree, node, tree[[pN]])
	}
	dist	
}


# Potential function
# list(N=-1, pN=-1, cNs=integer(0), x=NA, y=NA, w=NA, m=NA, f=NA, F=NA)
potential <- function(tree, params, n, dw) {
	pp = 0
	nn = n
	alpha = params["alpha"]
	beta = params["beta"]
	repeat {
		node = tree[[nn]]
		pnn = tree[[nn]]$pN
		if (pnn <= 0) {
			# got to root
			break
		}
		d = dist.to.node(tree, node)
		w = node$w
		m = node$m
		pp = pp + m^beta * ((w+dw)^alpha - w^alpha) * d
		nn = pnn
	}
	pp
}



################################################################################
# Filling in costs of veins: weights, bends, accumulation

fillin.weights <- function(tree) {
	Ns = collect.postorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		w = 1
		nc = length(node$cNs)
		if (nc > 0) {
			for (ch in 1:nc) {
				cN = node$cNs[ch]
				w = w + tree[[cN]]$w
			}
		}
		tree[[N]]$w = w
	}
	tree
}


bendcost <- function(pdx, pdy, dx, dy) {
	cos.theta = 0
	dotp = pdx*dx + pdy*dy
	if (dotp > 0) {
		dist = sqrt((pdx^2 + pdy^2) * (dx^2 + dy^2))
		cos.theta = dotp / dist
	}
	1 / cos.theta
}





fillin.bends <- function(tree) {
	Ns = collect.preorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		m = 1
		pN = node$pN
		if (pN > 0) {
			x = node$x
			y = node$y
			px = tree[[pN]]$x
			py = tree[[pN]]$y
			# parent coords of root in case no grandparent of N
			ppx = 0
			ppy = -1
			ppN = tree[[pN]]$pN
			if (ppN > 0) {
				ppx = tree[[ppN]]$x
				ppy = tree[[ppN]]$y
			}
			pdx = px - ppx
			pdy = py - ppy
			dx = x - px
			dy = y - py
			b = bendcost(pdx, pdy, dx, dy)
			m = tree[[pN]]$m * b
		}
		tree[[N]]$m = m
	}
	tree
}




fillin.vein.costs <- function(tree, alpha, beta) {
	nnodes = length(tree)
	for (N in 2:nnodes) {
		node = tree[[N]]
		pN = node$pN
		stopifnot(pN > 0)
		dx = node$x - tree[[pN]]$x
		dy = node$y - tree[[pN]]$y
		tree[[N]]$f = node$m^beta * node$w^alpha * sqrt(dx^2 + dy^2)
	}
	Ns = collect.postorder(tree)
	for (N in Ns) {
		node = tree[[N]]
		F = tree[[N]]$f
		nc = length(node$cNs)
		if (nc > 0) {
			for (ch in 1:nc) {
				cN = node$cNs[ch]
				F = F + tree[[cN]]$F
			}
		}
		tree[[N]]$F = F
	}
	tree
}



fillin.all.costs <- function(tree, params) {
	tree = fillin.weights(tree)
	tree = fillin.bends(tree)
	tree = fillin.vein.costs(tree, params["alpha"], params["beta"])
	tree
}


################################################################################
# making grid versions of the tree for both display, and finding
# cells next to border to (possibly) add.

# Note that grid is (row, column) so (y,x) and image has rows increasing
# upwards, unlike matrix
# root is at (0,0) in tree, (1-miny, 1-minx) in grid


tree.bbox <- function(tree) {
	minx = Inf
	maxx = -Inf
	miny = Inf
	maxy = -Inf
	nnodes = length(tree)
	for (n in 1:nnodes) {
		minx = min(minx, tree[[n]]$x)
		maxx = max(maxx, tree[[n]]$x)
		miny = min(miny, tree[[n]]$y)
		maxy = max(maxy, tree[[n]]$y)
	}
	list(minx=minx, maxx=maxx, miny=miny, maxy=maxy, w = 1 + maxx - minx, h = 1 + maxy - miny) 
}


x.y.to.N <- function(grid, bbox, x, y) {
	xx = 1 + x - bbox$minx
	yy = 1 + y - bbox$miny
	n = -1
	if (xx >= 1 & xx <= bbox$w & yy >= 1 & yy <= bbox$h )
	{
		n = grid[yy,xx]
	}
	n
}



tree.as.grid.nodenums <- function(tree, bbox) {
	grid = matrix(-1, nrow=bbox$h, ncol=bbox$w)
	nnodes = length(tree)
	for (n in 1:nnodes) {
		x = 1 + tree[[n]]$x - bbox$minx
		y = 1 + tree[[n]]$y - bbox$miny
		grid[y, x] = tree[[n]]$N
	}
	grid
}



tree.as.grid.Fs <- function(tree, bbox) {
	grid = matrix(0, nrow=bbox$h, ncol=bbox$w)
	nnodes = length(tree)
	maxF = tree[[1]]$F
	for (n in 1:nnodes) {
		x = 1 + tree[[n]]$x - bbox$minx
		y = 1 + tree[[n]]$y - bbox$miny
		grid[y, x] = tree[[n]]$F / maxF
	}
	grid
}



################################################################################
# Vein optimisation


try.reparent.to.x.y <- function(tree, params, grid, bbox, n, x, y) {
	tree2 = tree
	qN = x.y.to.N(grid, bbox, x, y)
	if (qN > 0) {  
		# if qN is in tree, check it is not n, nor an ancestor of n
		loop = FALSE
		nn = qN
		repeat {
			if (nn == n) {
				loop = TRUE
				break
			}
			nn = tree[[nn]]$pN
			if (nn <= 0) {
				# got to root
				break
			}
		}
		if (!loop) {
			tree2 = reparent.node(tree, n, qN)
			tree2 = fillin.all.costs(tree2, params)
		}  
	} 
	tree2
}




tryto.improve.parent <- function(tree, params, grid, bbox, n) {
	node = tree[[n]]
	pN = node$pN
	best.new.tree = tree
	x = tree[[n]]$x
	y = tree[[n]]$y
	pN = tree[[n]]$pN
	
	stopifnot(is.numeric(pN))
	stopifnot(pN>0)
	
	stopifnot(is.numeric(pN))
	stopifnot(pN > 0)
	px = tree[[pN]]$x
	py = tree[[pN]]$y
	minx = min(x, px) - 1
	maxx = max(x, px) + 1
	miny = min(y, py) - 1
	maxy = max(y, py) + 1
	for (xx in minx:maxx) {
		for (yy in miny:maxy) {
			if ((abs(xx-x) <= 1 & abs(yy-y) <= 1)  |  (abs(xx-px) <= 1 & abs(yy-py) <= 1)) {
				tree2 = try.reparent.to.x.y(tree, params, grid, bbox, n, xx, yy)
				if (tree2[[1]]$F < best.new.tree[[1]]$F) {
					best.new.tree = tree2
				}
			}
		}
	}
	best.new.tree
}



tryto.improve.veins <- function(tree, params, grid, bbox) {
	nnodes = length(tree)
	for (n in 2:nnodes) {
		tree = tryto.improve.parent(tree, params, grid, bbox, n)
	}
	tree
}



optimise.veins <- function(tree, params) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.nodenums(tree, bbox)
	
	repeat {
		F1 = tree[[1]]$F
		tree = tryto.improve.veins(tree, params, grid, bbox)
		tree = fillin.all.costs(tree, params)
		F2 = tree[[1]]$F
		if (F2 == F1) break
	}
	tree
}




################################################################################
# Adding cells at border

best.parent <- function(tree, grid, bbox, params, x, y) {
	pNs = integer(0)
	for (xx in (x-1):(x+1)) {
		for (yy in (y-1):(y+1)) {
			pN = x.y.to.N(grid, bbox, xx, yy)
			if (pN > 0) {
				# (xx,yy) is in tree
				pNs = c(pNs, pN)
			}
		}
	}
	bestp = -1
	bestcost = Inf
	if (length(pNs) > 0) {
		beta = params["beta"]
		besti = -1
		for (i in 1:length(pNs)) {
			pn = pNs[i]
			cost = potential(tree, params, pn, 1)
			px = tree[[pn]]$x
			py = tree[[pn]]$y
			ppn = tree[[pn]]$pN
			ppx = Stalk.x
			ppy = Stalk.y # in case pn is root
			if (ppn > 0) {
				ppx = tree[[ppn]]$x
				ppy = tree[[ppn]]$y
			}
			dx = x-px
			dy = y-py
			b = bendcost(px-ppx, py-ppy, dx, dy)
			m = tree[[pn]]$m
			d = sqrt(dx*dx+dy*dy)
			cost = cost + (m*b)^beta * d
			if (cost < bestcost) {
				bestcost = cost
				besti = i
			}
		}
		if (besti > 0) {
			bestp = pNs[besti]
		}
	}
	list(p=bestp, cost=bestcost, x=x, y=y)
}



list.cells.to.add <- function(tree, params) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.nodenums(tree, bbox)
	new.tree = tree
	minx = bbox$minx - 1
	maxx = bbox$maxx + 1
	miny = bbox$miny - 1  
	maxy = bbox$maxy + 1
	cells = list()
	for (x in minx:maxx) {
		for (y in miny:maxy) {
			m = x.y.to.N(grid, bbox, x, y)
			if (m < 0) {
				# (x,y) is not in tree
				best = best.parent(tree, grid, bbox, params, x, y)
				if ((best$p > 0)) {
					cells = c(cells, list(best))
				}
			}
		}
	}
	mincost = Inf
	if (length(cells) > 0) {
		for (i in 1:length(cells)) {
			if (cells[[i]]$cost < mincost) {
				mincost = cells[[i]]$cost
			}
		}		
		cheapestcells = list()
		for (i in 1:length(cells)) {
			if (cells[[i]]$cost < mincost + 0.00001) {
				cheapestcells = c(cheapestcells, list(cells[[i]]))
			}
		}
		cells = cheapestcells
	}
	cells
}


add.cells.via.list <- function(tree, params) {
	cells = list.cells.to.add(tree, params)
	bbox = tree.bbox(tree)
	new.tree = tree
	numadded = 0
	cost = Inf
	if (length(cells) > 0) {
		cost = cells[[1]]$cost
		for (i in 1:length(cells)) {
			if (cells[[i]]$cost < params["epsilon"]) {
				new.tree = add.child(new.tree, cells[[i]]$p,  cells[[i]]$x,  cells[[i]]$y)
				numadded = numadded + 1
			}
		}		
	}
	list(new.tree=new.tree, added=numadded, cost=cost)
}



################################################################################
# Display, as image and text

tree.as.image <- function(tree, h, w, rooth) {
	bbox = tree.bbox(tree)
	grid = tree.as.grid.Fs(tree, bbox)
	plotgrid = matrix(0, nrow=h, ncol=w)
	x = 1:w
	y = 1:h
	offx = (w-1)/2
	offy = rooth
	for (r in 1:bbox$h) {
		for (c in 1:bbox$w) {
			xx = c + offx + bbox$minx
			yy = r + offy + bbox$miny
			if (yy >= 1  &&  yy <= h  &&  xx >= 1  &&  xx <= w) {
				plotgrid[yy, xx] = grid[r,c]
			}
		}
	}
	z = t(plotgrid)^.33
	nc = 100
	clrs = rev(hcl.colors(nc, palette = "viridis", alpha = NULL, rev = FALSE, fixup = TRUE))
	clrs = c("#f8f8f8", clrs)
	image(x, y, z, col=clrs, axes=FALSE, ann=FALSE, xlim=c(0,w)+0.5, ylim=c(0,h)+0.5)
	nnodes = length(tree)
	for (n in 2:nnodes) {
		node = tree[[n]]
		pnode = tree[[ node$pN ]]
		lines(x=1 + offx+c(pnode$x, node$x), y=1 + offy+c(pnode$y, node$y), lwd = 0.5*sqrt(node$w))
	}
}



tree.as.drawing <- function(tree) {
	bbox = tree.bbox(tree)
	nnodes = length(tree)
	ddd = matrix(0, nrow=nnodes, ncol=6)
	colnames(ddd) = c("x1", "x2", "y1", "y2", "w", "F")
	ddd[1,] = c(0, 0, -1, 0, tree[[1]]$w, tree[[1]]$F)
	for (n in 2:nnodes) {
		node = tree[[n]]
		pnode = tree[[ node$pN ]]
		ddd[n,] = c(pnode$x, node$x, pnode$y, node$y, node$w, node$F)
	}
  ddd
}



tree.nodes.as.matrix <- function(tree) {
	nnodes = length(tree)
	mm = matrix(0, nrow=nnodes, ncol=8)
	colnames(mm) = c("N","pN","x","y","W","m","f","F")
	for (n in 1:nnodes) {
		node = tree[[n]]
		mm[n,] = c(node$N, node$pN, node$x, node$y, node$w, node$m, node$f, node$F)
	} 
	mm
}


print.nodes <- function(tree) {
	mm = tree.nodes.as.matrix(tree)
	print(mm)
}



save.image <- function(tree, params, h, w, rooth) {
	fname = paste(as.character(100*params), collapse='-')
	fname = paste0(image.dir, fname, ".png")
	png(filename = fname, width = 20*w, height = 20*h, units = "px", pointsize = 12, bg = "white")
	par(mar=c(0,0,0,0))
	tree.as.image(tree, h, w, rooth)
	dev.off()
}


save.drawing <- function(tree, params, h, w, rooth) {
	fname = paste(as.character(100*params), collapse='-')
	fname = paste0(image.dir, fname, ".txt")
	ddd = tree.as.drawing(tree)
	write.table(ddd, file=fname, quote=FALSE, row.names = FALSE)
}


run.and.display <- function(params, h, w, root.y) {
	tree = make.root()
	repeat {
		growth = add.cells.via.list(tree, params)
		tree = growth$new.tree
		tree = fillin.all.costs(tree, params)
		
		cat("added", growth$added, "at cost", growth$cost, "to make", length(tree), "cells\n")
		if (growth$added == 0) {
			break
		}
		tree = optimise.veins(tree, params)
		tree.as.image(tree, h, w, root.y)
	}
	bbox = tree.bbox(tree)
	H = bbox$h + 2
	W = bbox$w + 2
	RootY = 1 - bbox$miny
	save.image(tree, params, H, W, RootY)
	save.drawing(tree, params, H, W, RootY)
	tree	
}


################################################################################
# examples - alpha, beta, epsilon  and image height, width, root position

example.55.100.50 <- function() {
	params = c(alpha = .55, beta = 1.00, epsilon = 5)
	run.and.display(params, 41, 29, 1)
}

example.55.70.50 <- function() {
	params = c(alpha = .55, beta = 0.70, epsilon = 5)
	run.and.display(params, 50, 39, 2)
}

example.55.30.40 <- function() {
	params = c(alpha = .55, beta = 0.30, epsilon = 4)
	run.and.display(params, 40, 29, 9)
}


example.95.150.150 <- function() {
	params = c(alpha = .95, beta = 1.50, epsilon = 15)
	run.and.display(params, 21, 23, 1)
}

example.95.200.200 <- function() {
	params = c(alpha = .95, beta = 2.00, epsilon = 20)
	run.and.display(params, 31, 31, 2)
}

example.95.1.150 <- function() {
	params = c(alpha = .95, beta = 0.01, epsilon = 13)
	run.and.display(params, 31, 31, 15)
}


example.55.1.50 <- function() {
	params = c(alpha = .55, beta = 0.01, epsilon = 5)
	run.and.display(params, 45, 45, 22)
}



little.example <- function() {
	params = c(alpha = .7, beta = .7, epsilon = 4)
	run.and.display(params, 15, 15, 1)
}



maple.example <- function() {
	params = c(alpha = .68, beta = .38, epsilon = 5)
	run.and.display(params, 21, 19, 4)
}

example.70.50.70 <- function() {
	params = c(alpha = .70, beta = .50, epsilon = 7)
	run.and.display(params, 35, 31, 6)
}



tiny.bendcost.example <- function() {
	params = c(alpha = .7, beta = 0.01, epsilon = 6.5)
	run.and.display(params, 55, 55, 27)
}




################################################################################
# top level

Stalk.x = 0
Stalk.y = -1


#little.tree = little.example()
maple.tree = maple.example()
#tree.55.100.50 = example.55.100.50()
#tree.55.70.50 = example.55.70.50()
#tree.95.150.150 = example.95.150.150()
#tree.55.30.40 = example.55.30.40()

#nbt = tiny.bendcost.example()

#example.95.1.150()
#example.55.1.50()

#example.70.50.70()

#example.95.200.200()









